@@ -1,6 +1,5 @@
 r""" ccp_abc.py - Parse, Query, Build, and Modify IOS-style configurations
-     Copyright (C) 2022      David Michael Pennington
-     Copyright (C) 2021      David Michael Pennington
+     Copyright (C) 2021-2022 David Michael Pennington
      Copyright (C) 2020-2021 David Michael Pennington at Cisco Systems
      Copyright (C) 2019      David Michael Pennington at ThousandEyes
      Copyright (C) 2014-2019 David Michael Pennington at Samsung Data Services
@@ -18,40 +17,51 @@ r""" ccp_abc.py - Parse, Query, Build, and Modify IOS-style configurations
      mike [~at~] pennington [/dot\] net
 """
 
+from difflib import get_close_matches
 from operator import methodcaller
 from abc import ABCMeta
+import warnings
 import inspect
 import re
+
 from ciscoconfparse.ccp_util import junos_unsupported
 from loguru import logger
 
-##
-##-------------  Config Line ABC
-##
+DEFAULT_TEXT = "__undefined__"
+
+#
+# -------------  Config Line ABC
+#
 class BaseCfgLine(metaclass=ABCMeta):
     # deprecating py2.foo metaclass syntax in version 1.6.8...
-    #__metaclass__ = ABCMeta
-    def __init__(self, text="__undefined__", comment_delimiter="!"):
+    # __metaclass__ = ABCMeta
+    def __init__(self, text=DEFAULT_TEXT, comment_delimiter="!"):
         """Accept an IOS line number and initialize family relationship
         attributes"""
         self.comment_delimiter = comment_delimiter
         self._text = text
+        self._uncfgtext_to_be_deprecated = ""
         self.linenum = -1
         self.parent = self
         self.child_indent = 0
         self.is_comment = None
         self.children = list()
-        self.oldest_ancestor = False
-        self.indent = 0  # Whitespace indentation on the object
+        self.indent = 0      # Whitespace indentation on the object
         self.confobj = None  # Reference to the list object which owns it
-        self.feature = ""  # Major feature description
-        self.feature_param1 = ""  # Parameter1 of the feature
-        self.feature_param2 = ""  # Parameter2 of the feature (if req'd)
+        self.feature = ""    # Major feature description
+        self.blank_line_keep = False  # set blank_line_keep True for banner / macro
         self.set_comment_bool()
 
-        self._diff_id = None
+        self._line_id = None
         self.diff_rendered = None
-        self.diff_word = ""
+        self.diff_word = ""  # diff_word: 'keep', 'remove', 'unchanged', 'add'
+        self.diff_side = ""  # diff_side: 'before', 'after' or ''
+
+        # FIXME
+        #   Bypass @text.setter method for now...  @text.setter writes to
+        #   self._text, but currently children do not associate correctly if
+        #   @text.setter is used as-is...
+        # self.text = text
 
     # On BaseCfgLine()
     def __repr__(self):
@@ -114,7 +124,6 @@ class BaseCfgLine(metaclass=ABCMeta):
         self.is_comment = retval
         return retval
 
-
     # On BaseCfgLine()
     @property
     def index(self):
@@ -122,48 +131,56 @@ class BaseCfgLine(metaclass=ABCMeta):
         return self.linenum
 
     # On BaseCfgLine()
-    def calculate_diff_id(self):
+    def calculate_line_id(self):
         """
-        Calculate and return this object's integer diff_id
+        Calculate and return an integer line_id for BaseCfgLine()
 
-        Do NOT cache this value.  It may need to be recalculated
-        if self.text changes.
+        line_id is used to build a numerical identity for a given
+        BaseCfgLine().
+
+        Do NOT cache this value.  It must be recalculated when self._text
+        changes.
         """
-
-        _diff_id = hash(" " * self.indent + " ".join(self.text.strip().split()))
-        return _diff_id
+        indent = self.indent
+        _line_id = hash(" " * indent + " ".join(self.text.strip().split()))
+        return _line_id
 
     # On BaseCfgLine()
     @property
     def diff_id_list(self):
         """
-        Return a list of integers as a unique diff identifier.
+        Return a list of integers as a context-sensitive diff identifier.
 
-        Be aware that object id integers are NOT the same between script runs.
+        The returned value includes line_id of all parents.  The oldest
+        ancestor / parent line_id is last in the returned list of line_id
+        hash values.
+
+        object id integers are NOT the same between script runs.
         """
-        retval = list()
+        retval = []
         len_geneology = len(self.geneology)
 
-        #while finished is not True:
         for idx, obj in enumerate(self.geneology):
 
-            obj._diff_id = obj.calculate_diff_id()
+            obj._line_id = obj.calculate_line_id()
 
             # idx = 0 is the oldest ancestor
-            if idx==0:
+            if idx == 0:
                 # This object is NOT a child
-                assert obj.indent == 0  # FIXME 2022-03-07
-                retval.insert(0, obj._diff_id)
-                # FIXME -> ORIGINAL obj = obj.parent
-                obj.parent = obj
-                #finished = True
+                assert obj.indent == 0
+                retval.insert(0, obj._line_id)
 
             elif idx <= len_geneology - 1:
                 # This object is a child of self.parent
-                assert obj.indent > 0  # FIXME 2022-03-07
-                retval.insert(0, obj._diff_id)
-                continue
+                if False:
+                    assert obj.indent > 0
+                retval.insert(0, obj._line_id)
 
+        # retval usually looks like this (example with a single parent obj):
+        #
+        #                          [-1387406312585020591, 3965133112392387338]
+        #  root / oldest _line_id:                        ^^^^^^^^^^^^^^^^^^^
+        #  child object _line_id:   ^^^^^^^^^^^^^^^^^^^^
         return retval
 
     # On BaseCfgLine()
@@ -174,25 +191,71 @@ class BaseCfgLine(metaclass=ABCMeta):
     # On BaseCfgLine()
     @text.setter
     def text(self, newtext=None):
+        # FIXME - children do not associate correctly if this is used as-is...
         assert isinstance(newtext, str)
+
+        # escape braces since single braces could be misunderstood as
+        # f-string or string.format() delimiters...
+        #
+        # Sometimes brace escaping is not required... we need better fencing
+        # around safe_escape_curly_braces()
+        if False:
+            newtext = self.safe_escape_curly_braces(newtext)
+
         previous_text = self._text
         self._text = newtext
-        # FIXME - for the sake of speed, only calculate_diff_id() when a
-        #         diff is requested.  I have not added the logic
-        #         for conditional diff calculation yet.
+
         if previous_text != newtext:
-            self._diff_id = self.calculate_diff_id()
+            self._line_id = self.calculate_line_id()
+
+            # After changing text we need to trigger
+            #     ConfigObjs.bootstrap_from_text()...
+            #
+            # If we take False out, it would run _bootstrap_from_text()
+            #     too often during initial configuration parsing...
+            #
+            # I put this here for some debugging but it shouldn't be
+            #     used in a general config parsing case.
+            #
+            # <-- Begin Removed -->
+            #if self.confobj is not None:
+            #    self.confobj._bootstrap_from_text()
+            # <-- End Removed -->
+
+    def safe_escape_curly_braces(self, text):
+        """
+        Escape curly braces in strings since they could be misunderstood as
+        f-string or string.format() delimiters...
+
+        If BaseCfgLine receives line with curly-braces, this method can
+        escape the curly braces so they are not mis-interpreted as python
+        string formatting delimiters.
+        """
+        # Bypass escaping curly braces if there aren't any...
+        if not ("{" in text) and not ("}" in text):
+            return text
+
+        assert ("{" in text) or ("}" in text)
+
+        # Skip escaping curly braces if text already has double curly braces...
+        if ("{{" in text) or ("}}" in text):
+            return text
+
+        text = text.replace("{", "{{")
+        text = text.replace("}", "}}")
+        return text
+
 
     # On BaseCfgLine()
     @property
-    def diff_id(self):
-        return self._diff_id
+    def line_id(self):
+        return self._line_id
 
     # On BaseCfgLine()
-    @diff_id.setter
-    def diff_id(self, value=None):
+    @line_id.setter
+    def line_id(self, value=None):
         assert isinstance(value, int)
-        self._diff_id = value
+        self._line_id = value
 
     # On BaseCfgLine()
     @property
@@ -203,18 +266,22 @@ class BaseCfgLine(metaclass=ABCMeta):
     @property
     def hash_children(self):
         """Return a unique hash of all children (if the number of children > 0)"""
-        if len(self.children) > 0:
-            return hash(tuple(self.children))
+        if len(self.all_children) > 0:
+            return hash(tuple(self.all_children))
         else:
-            return 0
+            return hash(())
 
     # On BaseCfgLine()
     @property
     def family_endpoint(self):
-        if self.children == []:
-            return 0
+        assert isinstance(self.all_children, list)
+        if self.all_children == []:
+            # CHANGED on 2022-04-01... PREVIOUS_VALUE="return 0"
+            # CHANGED on 2022-04-01... NEW_VALUE="self.linenum"
+            return self.linenum
         else:
-            return self.children[-1].linenum
+            return self.all_children[-1].linenum
+
 
     # On BaseCfgLine()
     @property
@@ -274,7 +341,10 @@ class BaseCfgLine(metaclass=ABCMeta):
     # On BaseCfgLine()
     @property
     def is_config_line(self):
-        """Return a boolean for whether this is a config statement; returns False if this object is a blank line, or a comment"""
+        """
+        Return a boolean for whether this is a config statement; returns False
+        if this object is a blank line, or a comment.
+        """
         if len(self.text.strip()) > 0 and not self.is_comment:
             return True
         return False
@@ -317,13 +387,79 @@ class BaseCfgLine(metaclass=ABCMeta):
 
     # On BaseCfgLine()
     @junos_unsupported
-    def add_uncfgtext(self, unconftext):
-        """unconftext is defined during special method calls.  Do not assume it
-        is automatically populated."""
-        ## remove any preceeding "no "
-        conftext = re.sub(r"\s*no\s+", "", unconftext)
+    def add_uncfgtext(self, unconftext=None):
+        """
+        add_uncfgtext() is deprecated and will be removed.
+
+        .. code-block:: python
+           :emphasize-lines: 16
+           >>> # assume parse.find_objects() returned a value in obj below
+           >>> obj.text
+           ' no ip proxy-arp'
+           >>> obj.uncfgtext
+           ''
+           >>> obj.add_uncfgtext(" no ip proxy-arp")
+        """
+        assert isinstance(unconftext, str)
+        assert isinstance(self.text, str) and self.text != DEFAULT_TEXT
+
+        # adding a deprecation warning in version 1.7.0...
+        deprecation_warn_str = "add_uncfgtext() is deprecated and will be removed."
+        warnings.warn(deprecation_warn_str, DeprecationWarning)
+        ## remove any preceeding "no " from Cisco IOS commands...
+        conftext = re.sub(r"^(\s*)(no\s+)(\S.*)?$", "\3", unconftext)
         myindent = self.parent.child_indent
-        self.uncfgtext = myindent * " " + "no " + conftext
+
+        # write access to self.uncfgtext is not supported
+        self._uncfgtext_to_be_deprecated = myindent * " " + "no " + conftext
+
+    @property
+    def uncfgtext(self):
+        """
+        Return a 'best-effort' Cisco IOS-style config to remove this
+        configuration object.
+
+        This `uncfgtext` string should not be considered correct
+        in all Cisco IOS command unconfigure cases.
+        """
+        assert isinstance(self.text, str) and self.text != DEFAULT_TEXT
+
+        tmp = [ii.strip() for ii in self.text.split()]
+
+        # _uncfgtext_to_be_deprecated is normally set in add_uncfgtext()...
+        # deprecation warnings for _uncfgtext_to_be_deprecated were
+        # introduced in version 1.7.0...
+        if self._uncfgtext_to_be_deprecated != "":
+            # Officially, _uncfgtext_to_be_deprecated is not supported.
+            # This if-case is here for someone who may have set
+            # self.uncfgtext in older ciscoconfparse versions.
+            #
+            # After uncfgtext.setter (below) is removed, we can rip out
+            # this portion of the if-else logic...
+            deprecation_warn_str = "add_uncfgtext() is deprecated and will be removed."
+            warnings.warn(deprecation_warn_str, DeprecationWarning)
+            return self._uncfgtext_to_be_deprecated
+
+        # Once _uncfgtext_to_be_deprecated is removed, we can make this
+        # condition the first in this if-else logic...
+        elif tmp[0].lower()=="no":
+            assert len(tmp) > 1 # join() below only makes sense if len(tmp)>1
+            return self.indent * " " + " ".join(tmp[1:])
+
+        else:
+            return self.indent * " " + "no " + self.text.lstrip()
+
+    @uncfgtext.setter
+    def uncfgtext(self, value=""):
+        # Officially, _uncfgtext_to_be_deprecated is not supported. This
+        # setter is here for someone who may have set self.uncfgtext in older
+        # ciscoconfparse versions.
+        #
+        # This uncfgtext setter was added in version 1.7.0...
+        deprecation_warn_str = "setting uncfgtext is deprecated and will be removed."
+        warnings.warn(deprecation_warn_str, DeprecationWarning)
+
+        self._uncfgtext_to_be_deprecated = value
 
     # On BaseCfgLine()
     @junos_unsupported
@@ -424,9 +560,11 @@ class BaseCfgLine(metaclass=ABCMeta):
 
     # On BaseCfgLine()
     @junos_unsupported
-    def insert_before(self, insertstr):
-        """Usage:
-            confobj.insert_before('! insert text before this confobj')"""
+    def insert_before(self, insertstr=""):
+        """
+        Usage:
+        confobj.insert_before('! insert text before this confobj')
+        """
         retval = None
         calling_fn_index = 1
         calling_filename = inspect.stack()[calling_fn_index].filename
@@ -434,25 +572,30 @@ class BaseCfgLine(metaclass=ABCMeta):
         calling_lineno = inspect.stack()[calling_fn_index].lineno
         error =  "FATAL CALL: in {} line {} {}(insertstr='{}')".format(calling_filename, calling_lineno, calling_function, insertstr)
         if isinstance(insertstr, str) is True:
-            retval = self.confobj.insert_before(self, insertstr, atomic=False)
+            retval = self.confobj.insert_before(insertstr, atomic=False)
+
         elif isinstance(insertstr, BaseCfgLine) is True:
-            retval = self.confobj.insert_before(self, insertstr.text, atomic=False)
+            retval = self.confobj.insert_before(insertstr.text, atomic=False)
+
         else:
             raise ValueError(error)
+
         #retval = self.confobj.insert_after(self, insertstr, atomic=False)
         return retval
 
     # On BaseCfgLine()
     @junos_unsupported
-    def insert_after(self, insertstr):
+    def insert_after(self, insertstr=""):
         """Usage:
             confobj.insert_after('! insert text after this confobj')"""
+
         # Fail if insertstr is not the correct object type...
         #   only strings and *CfgLine() are allowed...
-        if not getattr(insertstr, "capitalize", False) and not isinstance(insertstr, 'BaseCfgLine'):
+        if not isinstance(insertstr, str) and not isinstance(insertstr, 'BaseCfgLine'):
             error = "Cannot insert object type - %s" % type(insertstr)
             logger.error(error)
             raise NotImplementedError(error)
+
         retval = None
         calling_fn_index = 1
         calling_filename = inspect.stack()[calling_fn_index].filename
@@ -461,12 +604,15 @@ class BaseCfgLine(metaclass=ABCMeta):
         error =  "FATAL CALL: in {} line {} {}(insertstr='{}')".format(calling_filename, calling_lineno, calling_function, insertstr)
         if self.confobj.debug >= 1:
             logger.debug("Inserting '{}' after '{}'".format(insertstr, self))
-        if getattr(insertstr, "capitalize", False):
+
+        if isinstance(insertstr, str) is True:
             # Handle insertion of a plain-text line
-            retval = self.confobj.insert_after(self, insertstr, atomic=False)
+            retval = self.confobj.insert_after(insertstr, atomic=False)
+
         elif isinstance(insertstr, 'BaseCfgLine'):
             # Handle insertion of a configuration line obj such as IOSCfgLine()
-            retval = self.confobj.insert_after(self, insertstr.text, atomic=False)
+            retval = self.confobj.insert_after(insertstr.text, atomic=False)
+
         else:
             logger.error(error)
             raise ValueError(error)
@@ -610,10 +756,12 @@ class BaseCfgLine(metaclass=ABCMeta):
             A string or python regular expression, which should replace the text matched by ``regex``.
         ignore_rgx : str
             A string or python regular expression; the replacement is skipped if :class:`~models_cisco.IOSCfgLine` text matches ``ignore_rgx``.  ``ignore_rgx`` defaults to None, which means no lines matching ``regex`` are skipped.
+
         Returns
         -------
         str
             The new text after replacement
+
         Examples
         --------
         This example illustrates how you can use
@@ -645,14 +793,17 @@ class BaseCfgLine(metaclass=ABCMeta):
         """
         # When replacing objects, check whether they should be deleted, or
         #   whether they are a comment
-        if ignore_rgx and re.search(ignore_rgx, self.text):
-            return self.text
-        retval = re.sub(regex, replacergx, self.text)
+        if ignore_rgx and re.search(ignore_rgx, self._text):
+            return self._text
+
+        retval = re.sub(regex, replacergx, self._text)
+
         # Delete empty lines
         if retval.strip() == "":
             self.delete()
             return
-        self.text = retval
+
+        self._text = retval
         self.set_comment_bool()
         return retval
 
